/* IZ 050705 {
UNDER DEVELOPMENT

A script object holds a text script that creates a process that can be started and stopped.
It creates a button as a gui for starting and stopping the script plus sliders for controlling parameters.
See more doc in "Scripts/About Scripts/the Script class"
}*/
Script : Model {
	// {
	classvar <>globalParentEnvir;
	classvar <>lastLoadedScripts; // list of all scripts from last opened folder
	classvar <globalScripts; 		/* dict (event) of scripts entered in globalParentEnvir 
			by addGlobalScript(globalname, script). Used to display these scripts in: */
	classvar <globalScriptsWindow; // display list of globally known scripts
	classvar <>midiConfig; // a MIDIConfig for convenient configuration - see autoMidiSetup
			// originally provided by Session, more convenient to store it here than to have 
			// to provide it all the time as argument.

	var <name;		// this name will be displayed on the gui button
	var <>path;		// the path from which the script was loaded
	var <>script;		// the script as text (can be edited and resaved)
	var <>envir;		// the environment within which the script will be evaluated
	var <>startFunc;	// Function put by script in envir.start.  Upon this.start, it
					// creates the playing process. Must return one of: Synth, 
					// Function (becomes Routine), Pattern (played by PatternPlayer)
	var <>stopFunc;	// put by script in envir.stop - defaults to 
					// { process.perform(*(envir.stopMessage ?? { process.stopMessage })) }
	var <>process;	// The process that is playing while this script is running.
					// Created from startFunc. Can be: Synth, Routine, PatternPlayer.
	var <inputs;		/* dict of busses which my process reads signals from by map or set In.ar
		Each association is: ctlname->linkedbus. Before this process starts, it notifies 
		each linked bus, so it allocates and sets / maps the ctlname to the bus index.
		When the process stops, it notifies each bus so it will free if no more processes
		are listening or writing to it */
	var <signalOutput; 	// A single Script that I am outputting my \out signal to 
	var <scalarOutputs;	// event of scripts that receive values from me via the method 'out'
	var <gui;			// control gui window. Only one is created
	var >guiBounds;	// rect of last closed guiWindow: for reopening on same place
	var <parameters;	// array of Parameters (for GUI and MIDI)
	var <midiResponder;// midiResponder for starting or stopping this script
	var <>isActive = false;		// whether midiResponder is active
	var <>id;			// object hash id if saved to file - used for reloading links
	/* ---------------------------------------------------------------------------------
	   following describe the type of Script based on the presence of keys in controlSpecs
	   and are used for coloring its main button and deciding can-receive-drag */
	var <hasBufferInput = false;
	var <hasAudioInput = false;
	var <hasAudioOutput = false;
	var <hasControlOutput = false;
	var <window; // stores window where my gui is - for saving and reloading sessions
	// ======= following are not yet fully implemented and are subject to changes =======
	var <snapshots;		// snapshots of full script states that can be saved on disc.
	var <buffers;		// buffer params->paths for save/
	// }
	*initClass {
		Class.initClassTree(Event);
		// make the usual patterns playable with default parent event
		globalParentEnvir = Event.parentEvents[\default].copy;
		// adding basic OSC driveability:
		Class.initClassTree(OSCresponder);
		OSCresponder(nil, 'Script', { | time, resp, msg |
			var tag;
//			{ this.changed(*msg) }.defer;
			#tag ... msg = msg;
			this.perform(*msg);
		}).add;
		globalScripts = ();
	}
	*new { | name, path, script, envir |
		^super.newCopyArgs(nil, name ? 'empty script', path, script, envir).init;
	}
	fromSpecs { | argName, argPath, argScript |
	// used by ScriptPatch
		this.name = argName;
		path = argPath;
		script = argScript;
		envir = nil;
		this.init;
	}
	init {
		envir = envir ?? {(
			start: #{Routine({ 5.do { "empty script".postln; 1.wait }}).play};
		)};
		envir.parent = globalParentEnvir;
		envir.script = this;
		inputs = ();
		scalarOutputs = ();
		signalOutput = nil;
		this.initScript;
	}
	initScript {
		var key, action, spec;
		if (this.isRunning) { this.stop; };
		this.close; 	// (close and) update GUIs 
		// compile the actual script text and evaluate the resulting function within the 
		// environment.  This sets up the script instance for starting and stopping:
		if (script.notNil) { envir.use(script.compile); };
		startFunc = envir[\start];
		if (startFunc.isNil) {
			warn("Script " ++ name.asString ++ " does not set the ~start field");
		};
		stopFunc = envir[\stop] ?? {
			{ process.perform(*(envir.stopMessage ?? {[\quit, { this.stopped; }]})); };
		};
		/* - set input-output type properties for linking and coloring 
		   - initialize parameter values, otherwise synths wont work properly */
		// first re-init in case any of these have been set by previous script contents: 
		hasBufferInput = hasAudioInput = hasAudioOutput = hasControlOutput = false;
		// initialize control spec variables and create MIDIResponderHolders:
		parameters = envir.controlSpecs.collect { |spec|
			key = spec[0];
			// set connection properties for drag+drop and connecting:  
			switch ( key,
				\bufnum, { hasBufferInput = true },
				\in, { hasAudioInput = true },
				\out, { if (name.asString[..1] == "k_") {
						hasControlOutput = true;
					}{
						hasAudioOutput = true;
					}
				}
			); // create parameter and set initial value in envir:
			Parameter(this, spec).reset;
		};
	}
	makeAction { | key, argAction |
		/* function to call when parameter is sent a value from gui or elsewhere
		This method is called by Parameter-init */
		^if (argAction.notNil) {
			{ | fullVal, normalizedVal |
				envir.use({ argAction.value(fullVal, normalizedVal) });
			};
		} {
			{ | fullVal |
				envir[key] = fullVal;
				if (process.isKindOf(Node)) { process.set(key, fullVal) };
			};
		};
	}
	fromSynthDesc { | synthDesc |
		// { clear this script and replace with script made from synthdesc 
		// Uses script already made in class SynthDescs instead of remaking the script
		// FIX ME: fromSynthDesc should be class method? }
		envir = nil;
		script = SynthDefs.currentScript;
		path = "./" ++ SynthDefs.selectedDef.name.asString;
		this.name_(SynthDefs.selectedDef.name);
		this.init;
	}
	test {	// a gui for safely testing a script - stops when the gui closes
		// create a gui + make yourself stop when the gui closes
		this.makeGui;
		gui.addDependant { | who, state |
			if (state == \closed) { this.stop }
		};
		if (Server.local.serverRunning.not) { Server.local.boot };
	}
	*loadRaw { | folder, file, envir |
		// load a file as raw code, as found from folder/file names in ScriptBrowser
		// interpret the file as is (do not create a script instance)
		// provide access to an environment. Default: currentParentEnvir
		^this.loadRawEnvir(ScriptBrowser.getScriptPath(folder, file), envir);
	}

	*loadRawEnvir { | path, envir |
		// load a file as is, without script instance
		// provide environment to the code as context. Default: globalParentEnvir
		var currentEnvirBackup, result;
		currentEnvirBackup = currentEnvironment;	// backup current environment
		envir = envir ? globalParentEnvir;		// use custom environment
		currentEnvironment = envir;				// (default is globalParentEnvir
		result = path.load;					// load within environment
		currentEnvironment = currentEnvirBackup;	// restore original current environment
		^result;		// also return result of evaluation for further use
	}

	*load { | path |
		// load all scripts that match this paths string
		^this.loadPaths(path.pathMatch);
	}
	*loadPaths { | paths |
		// load all scripts in an array of paths. Do not load folders
		^paths.select { |p| p.last != $/ }.collect { |p| this.new.load(p); }
	}
	load { | argPath |
		// load a script from file into this script, updating name and gui
		var file;
		path = argPath ? path;
		if (File.exists(path).not) {
			^Post << "Cannot load script from path: " << path << "\n";
		};
		file = File(path, "r");
		this.script = file.readAllString;
		file.close;
		this.name = path.splitext.first.basename;
		envir = nil;	// leaving old envir can have unwanted results!
		this.init;
	}
	openDialog {
		// load one script from path selected by user into his instance 
		CocoaDialog.getPaths({ | paths |
			this.load(paths.first)
		}, {
			"load cancelled".postln;
		})
	}
	*openFolderDialog {
		// load all scripts in folder selected by user and make new gui
		CocoaDialog.getPaths({ | paths |
			paths = paths.first.dirname;
			this.load(paths ++ "/*");
			this.makeGui(name: paths.basename);
		}, {
			"load cancelled".postln;
		})
	}
	name_ { | argName |
		name = argName.asSymbol;
		this.changed(\name, name);
	}
	start { | ... args | // args may be passed e.g. by MIDIResponder functions 
		// do not restart if started by a script whose output you receive
 		if (this.isRunning) { ^this };
		CmdPeriod.add(this);
		if (hasAudioInput) { this.getInputBus(\in) };
		inputs.keysValuesDo { | param, bus | bus.start(this); };
		if (signalOutput.notNil) { signalOutput.key.start };
		process = envir.use({startFunc.(*args)});
		process.onEnd({this.stopped}, envir);
		envir.process = process;
		this.changed(\started);
		^process;
	}
	stop { | ... args | // args may be passed e.g. by MIDIResponder functions
		if (this.isRunning) {
			envir.use({stopFunc.(*args)});
			if (process.isKindOf(Node)) { 
				this.changed(\released)
			}{
				this.stopped;
			};
		}
	}
	kill { process.free }
	resetParams {
		var key, action;
		envir[\controlSpecs].do { |s|
			key = s[0];
			if ([\in, \out, \bufnum].includes(s[0]).not) {
				// set each parameter except for in, out, bufnum to default
				if ((action = s[6]).isNil) {
				// use default action if no action is provided
					this.set(key, s[5] ?? {s[1]})
				}{
				// use custom action for setting if provided
					envir.use { action.(s[5]) };
				}
			}
		}
	}
	stopped {
		inputs.keysValuesDo { | param, bus | bus.stop(this); };
		process = nil;
		envir[\process] = nil;
		this.changed(\stopped);
	}
	toggle {
		if (this.isRunning) { this.stop } { this.start }
	}
	isRunning { ^process.notNil }
	cmdPeriod {
		CmdPeriod.remove(this);
		this.stopped;
	}
	/*  CONSTRUCTING START FUNCTIONS FOR FUNCTIONS, ENVIRONMENTS, PATTERNS  */
	routine { | func |
		// Routine made from a function
		envir[\start] = {
			Routine({
				envir.use(func);
				this.stopped;
			}).play(envir[\clock] ? SystemClock)
		}
	}
	playEnvir {
		// The envir Event: 
		envir[\start] = { Pbindf(envir).play }
	}
	asPbind { | ... keysValues |
		// Pbindf with the envir Event
		envir[\start] = { Pbindf(envir, *keysValues).play }
	}
	session { // Construct stop message to close session stored in process when 
		// script is stopped. More code may be added to the present method?
		envir[\stopMessage] = \remove;
	}
	/* { LISTENING TO BROADCAST MESSAGES FROM OTHER OBJECTS 
	Create a start function which adds me as dependant to object stored in ~sender, 
	reacting to messages whose second argument is equal to <message> } */
	listenTo { | action, sendersKey = \sender, messageKey = \vertex |
		~script.routine {
			var adapter;
			envir.use(envir[\startScore]);
			adapter = { | sender, message ... args |
				if (message == messageKey) {
					envir.use { action.(*args) };
				};
			};
			envir[sendersKey].addDependant(adapter);
			this.yield;
			envir[sendersKey].removeDependant(adapter);
			envir.use(envir[\stopScore]);
		};
		~stopMessage = \next;
	}
	makeGui {
		var label, slider, numbox, controlspec, views, dragsink, adapterEnvir, adapter;
		if (gui.notNil) { ^gui.front };
		gui = SCWindow(name.asString, this.guiBounds);
		gui.view.keyDownAction = { | me, char, mod, unicode, key |
			switch (char,
				$m, { this.activateMidi },
				$M, { this.deactivateMidi },
				$n, { this.deactivateMidi },	// (convenient synonym for $M key)
				$e, { this.edit },
				$g, { this.showControlGui },	// show control gui part
				$h, { this.hideControlGui },	// hide control gui part
 //				$\r, { this.toggle; },			// return key: stop if running / start if stopped
				$ , { this.toggle; },			// space key: stop if running / start if stopped
				$k, { this.kill; },	// free synth immediately
				$r, { this.resetParams }// , next line would break numbox keydown action
//				{ me.defaultKeyDownAction(char, mod, unicode, key) }
			);
		};
		gui.view.decorator = FlowLayout(gui.view.bounds, 3@3, 1@1);
		this.makeMainButtonAndDragView(gui);
		gui.view.decorator.nextLine;
		adapterEnvir = (
			scriptClosed: { gui.close },
			midiOn: { gui.view.background = Color(0.1, 0.4, 0.7, 0.5); },
			midiOff: { gui.view.background = Color(0.9, 0.9, 0.9, 0.5); }
		);
		adapter = { | sender, state, fullVal |
			adapterEnvir[state].value(fullVal);
		};
		this.addDependant(adapter);
		gui.onClose = {
			guiBounds = gui.bounds;
			gui.changed(\closed);	// notify dependant (menu) views
			gui.release;
			gui = nil;
			this.removeDependant(adapter);
		};
		parameters.do { |p| p.makeGui(gui, adapterEnvir)};
		gui.front;
	}
	guiBounds {
		^guiBounds ?? { this.defaultGuiBounds }
	}
	defaultGuiBounds {
		^Rect(450, 420,
			if (parameters.size == 0) { 200 } { 400 },
			1 + parameters.size * 21 + 6
		).fromTop
	}

	placeGui { |x = 0, y = 0 |
		// TODO: Implement placeGui (sort out when the default rect is made)
		// move the gui window so that left top corner is at x, y. Also works before
		// window opens
		guiBounds = this.guiBounds.moveTo(x, y).fromTop;
	}
	closeGui {
		if (gui.notNil) { gui.close };
	}
	toggleGui {
		if (gui.notNil) { gui.close } { this.makeGui }
	}
	makeMainButtonAndDragView { | argWindow |
		var button, adapter, dragview;
		button = SCButton(argWindow, Rect(0, 0, 145, 20))
		.font_(Font("Helvetica", 11))
		.action_({|me| [{ this.stop }, {this.start}]
			[me.value].value(this);
		})
		.keyDownAction_({ | view, char, modifiers, unicode, keycode |
			switch ( char,
//				$u, { this.removeOutput },		// unlink - NOT YET DONE!
				$e, { this.edit },				// open script file for editing
				$g, { this.showControlGui },	// show control gui part
				$h, { this.hideControlGui },	// hide control gui part
				$o, { this.class.openDialog },	// open new script file
				$R, { this.load },				// reload edited script
				$m, { this.bind2MIDIFromUser },	// make controllable by next incoming MIDI
				$\r, { this.toggle; },			// return key: stop if running / start if stopped
				$ , { this.toggle; },			// space key: stop if running / start if stopped
				$k, { this.kill; },	// free synth immediately
				$r, { this.resetParams },
				{ view.defaultKeyDownAction(char, modifiers, unicode, keycode); }
			);
		});
		this.setButtonStates(button, argWindow);
		dragview = this.makeDragView(argWindow);
		adapter = { | sender, state ... moreArgs |
			switch (state,
				\stopped, {
					this.setButtonStates(button, argWindow);
				},
				\started, {
					this.setButtonStates(button, argWindow);
				},
				\released, { this.setReleasedButtonStates(button, argWindow); },
				\test, { [sender, state, moreArgs].postln; },
				\wait4MIDI, { button.flashWhile(*moreArgs) },
				// change button name. not yet done. 
				\name, { this.setButtonStates(button, argWindow, this.isRunning.binaryValue); },
				\inputLink, { this.checkInputLinkIndicationColor(dragview); },
				\outputLink, { this.checkOutputLinkIndicationColor(dragview); },
				\scriptClosed, { this.setButtonStates(button, argWindow); }
			)
		};
		this.addDependant(adapter);
		button.value = this.isRunning.binaryValue;
	}
	checkInputLinkIndicationColor { | dragview |
		if (inputs.size == 0) {
			{ dragview.stringColor = Color.black; }.defer;
		} {
			{ dragview.stringColor = Color.red; }.defer;
		}
	}
	checkOutputLinkIndicationColor { | dragview |
		if ((signalOutput.isNil) and: { scalarOutputs.size == 0 }) {
			{ dragview.background = Color.clear; }.defer;
		} {
			{ dragview.background = Color(0.5, 0.5, 1.0); }.defer;
		}
	}
	makeDragView { | argWindow |
		var drag; // , acceptDragFunc;
		drag = SCDragBoth(argWindow, Rect(0, 0, 14, 20))
			.object_(this).string_("@")
			.canReceiveDragHandler = {
				this.canReceiveDragHandler;
			};

		// NOT COMPLETE YET
		drag.receiveDragHandler = { | me |
			var cd;
			cd = SCView.currentDrag;
			case
			{ cd.isKindOf(MIDIResponder)} {
				this.midiResponder =cd;
			}
			{ cd.isKindOf(String) } {
				this.load(cd);
			}
			{ cd.isKindOf(Buffer) } {
				this.set(\bufnum, cd.bufnum);
			}
			// from synth desc disabled - use the s key for storing synthdef-script in 
			// scripts instead
/*			{ cd.isKindOf(SynthDesc) } {
				this.fromSynthDesc(cd);
			}
*/			// connect non-audio output pattern globally to any other script
			{

//				[hasAudioOutput, hasControlOutput].postln;
			not(cd.hasAudioOutput or: cd.hasControlOutput)

			} {
			// connecting non-signal output (pattern or control or object-listening script) 
			// to any other script: 
			// at the moment a simple brute-force connection: make 
			// receiver available to sender for control by putting the receiver in the 
			// senders environment under 'out'. No update of relevant 
			// link variables (outputs, inputs) or of GUI is done for now.  
				Post << "Connecting writer" << cd << " to reader " << this << "\n";
				cd.envir[\out] = this;
//				cd.envir.postln;
			}

			{ hasAudioInput } { // connect fx script to audio instrument script
				if (cd.hasAudioOutput) {
					this.addInput(cd, \in, \out)
				}{
//		// more work needed here. for the moment default acts as if cd had Audio Output
//					"connecting non-audio output script to audio input ".warn;
//					this.addInput(cd, \in, \out)
					"connection of non-audio output to audio input not yet implemented".warn;
				}
			}
			{ true } {

			};
			me.object = this;
			me.string = "@";
		};
		^drag;
	}
	canReceiveDragHandler {
		var cd;
		cd = SCView.currentDrag;
		if (cd.isKindOf(MIDIResponder)) { ^true };
		^if ( cd.isKindOf(SynthDesc) or: { cd.isKindOf(String) }) { true
		} {
			if ( cd.isKindOf(Buffer) ) {
				hasBufferInput
			} {
				cd.hasControlOutput.not
				and: {
						not(hasAudioInput.not and: { cd.hasAudioOutput })
				} and: {
					this.doesNotContainCycles(cd);
				}
			}
		}
	}
	doesNotContainCycles { | argInput |
		if (argInput == this) { ^false };
		if (signalOutput.isNil) { ^true };
		^signalOutput.key.doesNotContainCycles(argInput);
	}
	makeMenu { | argWindow |
		var menu;
		menu = SCPopUpMenu(argWindow, Rect(0, 0, 14, 20))
			.items_(["---", "make gui", "edit", "reload", "open ...", "open folder ...",
				"save", "save as ...", "bind to MIDI ...", "save config"
			])
			.action_({ |me|
				[{},
				{ this.makeGui },
				{ this.edit },
				{ this.load },
				{ this.openDialog },
				{ this.class.openFolderDialog },
				{ this.save },
				{ this.saveAs },
				{ this.bind2MIDIFromUser(true) },	// react to just 1 key
//				{ this.bind2MIDIFromUser(false) },	// react to all keys on a channel
				{ this.saveConfig }
				][me.value].value;
				me.value = 0;
			});
	}
	setButtonStates { | button, window |
		var backcolor, oncolor, offcolor;
		if (window.isClosed) { ^nil };
		#backcolor, offcolor, oncolor  = case
			{ hasAudioInput } { [Color(0.85, 0.0, 0.25), Color.white, Color(0.3, 0.9, 1.0)] }
			// Audio input has precedence over buffer input in coloring
			{ hasBufferInput } { [Color(1.0, 0.9, 0.5), Color.black, Color(1, 0.2, 0.2)] }
			{ hasAudioOutput } { [Color(0.5, 0.75, 1.0), Color.black, Color(1, 0.2, 0.2)] }
			{ hasControlOutput } { [Color(0.5, 1.0, 0.75), Color.black, Color(1, 0.2, 0.2)] }
			// default: no signal i/o, no buffers
			{ true } { [Color(1.0, 1.0, 1.0), Color.black, Color(1, 0.2, 0.2)] };
		{	button.states_([
				[name, offcolor, backcolor],
				["[" ++ name.asString ++ "]", oncolor, backcolor]
			]).refresh;
			button.value = this.isRunning.binaryValue;
		}.defer
	}
	setReleasedButtonStates { | button, window |
		if (window.isClosed.not) {
			{
					button.states_([
					["*" ++ name.asString]++this.releasedColors,
					["*" ++ name.asString]++this.releasedColors
				]).refresh;
			}.defer;
		};
	}
	offColors { |color| ^[Color.black, color] }
	onColors { |color| ^[Color(1, 0.2, 0.2), color] }
	releasedColors { ^[Color.red, Color.grey(0.8, 0.5)] }
	midiResponderKeys {
		^parameters.keys.asArray.sort;
	}
	// TODO: Redo *edit and edit to avoid duplication of code
	*edit { | path |
		// edit file of a script. Use Document window if already open on this path
		var found;
		found = Document.allDocuments.detect { | d | d.path == path };
		if (found.isNil) {
			if (File.exists(path)) {
				found = Document.open(path);
				found.path = path;
			} {
				Post << "Script File not found: " << path; 
				^nil; 
			};
			found.setFont(Font("Monaco", 9), -1).syntaxColorize;
		}{
			found.front;
		};
		^found;
	}
	edit {
		// edit the file of this script and reload when saved
		var found, tempPath;
		// use tempPath as placeholder in case path is nil to avoid wrong match
		// or error in searching file under nil path
		tempPath = path ?? { "./" ++ name.asString };
		found = Document.allDocuments.detect { | d | d.path == tempPath };
		if (found.isNil) {
			if (File.exists(tempPath)) {
				found = Document.open(tempPath);
				found.path = tempPath;
			} {
				found = Document(name.asString, script);
			};
			found.setFont(Font("Monaco", 9), -1).syntaxColorize;
		}{
			found.front;
		};
		found.onClose = { | doc |	// on close: reload document
			if (doc.path.isNil) {
				("Script '" ++ doc.title ++ "' closed without saving. Script not loaded").postln;
			} {
				("reloading script: " ++ doc.path).postln;
				this.load(doc.path);
			}
		};
		^found;
	}
	hideControlGui {
		if (gui.notNil) { gui.resize(24, 200) }
	}
	showControlGui {
		if (gui.notNil) { gui.resize(1 + envir.controlSpecs.size * 21 + 6, 400) }
	}
	makeSetterAction { | key, controlspec |
		^{ | fullVal, normalizedVal |
			envir[key] = fullVal;
			if (process.isKindOf(Node)) { process.set(key, fullVal) };
		};
	}
	addSynth { | defname ... args |
		/* Create and return a synth which outputs to your input channel given by 
			parameter 'in'.  */
		^Synth(defname, [\out, envir.in ? 0] ++ args, process, \addBefore);
	}
	getInputBus { | inputKey = \in |
		/* Obtain or create a bus that will be bound to one of your parameters 
			Used for linking synths of scripts to each other. See addInput */
		var bus;
		bus = inputs[inputKey];
		if (bus.isNil) {
			bus = LinkedBus(if (inputKey == \in) { \audio } { \control });
			bus.addReader(this, inputKey);
			bus.start(this);
			inputs[inputKey] = bus;
		};
		^bus;
	}
	// TODO: Rewrite addInput, addOutput to distinguish 1->n and n->1 connections
	addInput { | source, input = \in, output = \out |
		/* { make parameter named by input arg read its input from the output of another 
		  Script, given by argument source. 
		If no other source is already writing to this parameter, create a LinkedBus and
		store it in inputs. Else if a LinkedBus already exists, set the output of that 
		bus to write to that bus. 
		If I am already running, the bus is allocated immediately and connected to the 
		inputs and the outputs } */
		var bus;
		bus = this.getInputBus(input);
		bus.addWriter(source, output);
		// if input links exist, link-dragview string color <- red, else black 
		this.changed(\inputLink);
		// set color of sources link-dragview 
		source.addOutput(this, input, output);
		if (this.isRunning) { bus.start(this); };
	}
	addOutput { | source, input = \in, output = \out |
		/* Request was made for my output to write to source at parameter named by input arg.
		If I have no audio or control output, then add source to my scalarOutputs at input. Else 
		set my signalOutput to [source, input]. */
		if (hasAudioOutput or: { hasControlOutput }) {
			signalOutput = source->input;
		}{
			scalarOutputs[source] = input;
		};
		this.changed(\outputLink); 		// set color of sources link-dragview
	}
	set { | param, val |
		/* set a scripts parameter to a value and update dependants (gui etc) */
		envir.put(param, val);	// store new value, in case of restart
		// if process is a Synth or Group, set its parameter on the server
		if (process.isKindOf(Node)) { process.set(param, val) };
		this.changed(param, val);	// update dependants (gui etc)
	}
	map { | param, index |
		/* make control rate link to accept input from scripts that output to this script 
		set param to value of index and if process is node, map its value to that index */
		if (process.isKindOf(Node)) { process.map(param, index) };
	}
	out { | ... paramValuePairs |
		/* send series of values to parameters of all scripts stored in outputs array */
		scalarOutputs.do { | out |
			paramValuePairs.pairDo { | p, v | out.set(p, v) }
		}
	}
	storeArgs { ^[name] }
	printOn { | stream | this.storeOn(stream); }
	close {
		// called by Session etc when I am removed for ever 
		////// TODO: review close method for Script
		// THESE MUST BE DONE TO DELETE OR CHANGE SCRIPTS!

		// remove all inputs and outputs, notify gui, notify other scripts for updating 
		// their guis, free busses, stop if needed ...	
		// see also init and initScript
		this.closeGui;
		if (this.isRunning) { this.stop };
		this.changed(\scriptClosed);
		this.changed(\outputLink);
		this.changed(\inputLink);
		this.changed(\stopped);
		this.deactivateMidi;
		// more must be done here
		this.disconnect; 	/// ??? 
	}
	disconnect {
		// some part of close 

	}
	// ------------------ BUFFERS
	loadBuffer { | pathOrName, param = 'bufnum' |
		// set a BufParameter's value to use a specific buffer given either by name or by path.
		this.set(
			param,
			if (pathOrName.isKindOf(Symbol)) {
				Samples[pathOrName] ? 0;
			}{
				Samples.default.loadBuffer(pathOrName).bufnum;
			};
		);
	}
	// ------------------ MIDI
	midiResponder_ { | argResponder |
		if (midiResponder.notNil) { midiResponder.remove };
		midiResponder = argResponder.copy;
		if (isActive) { midiResponder.install };
//		midiResponder.function_(this.makeNoteOnResponderFunction, this.makeNoteOffResponderFunction);
		midiResponder.function_(
			{ | port, chan, num, vel |
//				['noteon?', this, port, chan, num, vel].postln;
				if (vel > 0) { this.start } { this.stop }
			},{ | port, chan, num, vel |
//				['noteoFFFFFFFFFFFF?', this, port, chan, num, vel].postln;
				this.stop;
			}
		);
	}
	makeNoteOnResponderFunction {
	/* { make function for responder from envir[\midiFuncSource] 
		or default.
		Provide script (this) through closure scope. 
	} */
		^("{ | script | " ++ this.getNoteOnFuncSource ++ "}"
		).interpret.(this);
	}
	getNoteOnFuncSource {
		var source;
		if ((source = envir[\noteOnFuncSource]).notNil) { ^source };
		^
"{|src,ch,num,vel|\
	['on', src, ch, num, vel].postln;
	if(vel==0) {\
		script.stop(num,vel)\
	}{\
		script.start(num, vel)\
	}\
}"
	}
	makeNoteOffResponderFunction {
		/* { make function for responder from envir[\midiFuncSource] 
			or default.
			Provide script (this) through closure scope. 
		} */
		^("{ | script | " ++ this.getNoteOnFuncSource ++ "}"
		).interpret.(this);
	}
	getNoteOffFuncSource {
		var source;
		if ((source = envir[\noteOffFuncSource]).notNil) { ^source };
		^
"{|src,ch,num,vel|\
	['off', src, ch, num, vel].postln;
		script.stop(num,vel)\
}"
	}
	activateMidi {
//		Post << name << ": ACTIVATING MIDI\n";
		if (midiResponder.notNil) {
			midiResponder.install;
			isActive = true;
			this.changed(\midiOn);
		};
		parameters.do {|p| p.activateMidi };
	}
	deactivateMidi {
//		Post << name << ": DEACTIVATING MIDI\n";
		if (midiResponder.notNil) {
			midiResponder.remove;
			isActive = false;
			this.changed(\midiOff);
		};
		parameters.do {|p| p.deactivateMidi };
	}
	bindMidi { | row = 0, column = 0 ... itemsToBind |
		// bind one or more parameters with responders copied from your 
		// midiConfig class variable holding a MIDIConfig.
		var item;
//		['------------------', row, column].postln;
		item = midiConfig.at(row, column);
		if (itemsToBind.size == 0) {
			itemsToBind = parameters.select { |p| p.canAutoMidiSetup }
		}{
			itemsToBind = itemsToBind.collect { |p|
				if (p == \script) {
					this
				}{
					parameters.detect { |par| par.name == p }
				}
			}
		};
		itemsToBind.do { |p|
			p.midiResponder = item.responder;
			item = item.nextItem;
		}
	}
	// ------------------ GLOBAL SCRIPT ITEMS
	*addGlobalScript { | globalName, script |
		// { Add and remove scripts inside the globalScripts classvar for global access
		//  Update display of global scripts on globalScriptsWindow } */
		globalScripts[globalName] = script;
		{this.changed(\addGlobalScript, globalName, script);}.defer;
	}
	*removeGlobalScript { | globalName |
		var script;
		script = globalScripts[globalName];
		if (script.notNil) { script.stop };
		globalScripts.removeAt(globalName);
		{this.changed(\removeGlobalScript, globalName);}.defer;
	}
	*makeGlobalsGui {
		// (at this stage:) Make window displaying the global scripts
		if (globalScriptsWindow.notNil) { ^globalScriptsWindow.front; };
		^this.makeGlobalsWindow;
	}
	*makeGlobalsWindow {
		var scriptList, adapter;
		globalScriptsWindow = SCWindow.new("Global Scripts", Rect(0, 420, 150, 150).fromTop);
		globalScriptsWindow.onClose = { globalScriptsWindow = nil };
		scriptList = SCListView(globalScriptsWindow, Rect(3, 3, 144, 144))
			.resize_(5)
			.font_(Font("Helvetica", 10));
		/* ( no action defined yet when selecing an item. This is just a passive display) */
		adapter = { | sender, state |
			switch (state,
				\removeGlobalScript,
					{ scriptList.items = globalScripts.keys.asArray.sort; },
				\addGlobalScript,
					{ scriptList.items = globalScripts.keys.asArray.sort; }
			)
		};
		this.addDependant(adapter);
		scriptList.onClose = { this.removeDependant(adapter) };
		{this.changed;}.defer;
		^globalScriptsWindow.front;
	}
// { REMOTE SCRIPT CONTROL: Methods for creating and controlling scripts via OSC
//	all these methods use the prefix r } */
	*rAdd { | globalName, folder, scriptName |
		// create new script and store it in globalParentEnvir under globalName
		this.addGlobalScript(globalName, ScriptBrowser.load(folder, scriptName));
	}
	*rRemove { | globalName |
		this.removeGlobalScript(globalName);
	}
	*rLoad { | globalName, folder, scriptName |
		// load script into an existing script (thereby replacing the previous
		// script loaded in that instance)
		globalScripts[globalName].load(ScriptBrowser.at(folder, scriptName));
		{ this.changed(\loadScript, globalName, folder, scriptName); }.defer;
	}
	*rStart { | globalName |
		{
			globalScripts[globalName].start;
			this.changed(\start, globalName);
		}.defer; // Defer needed for GUI actions from OSC
	}
	*rStop { | globalName |
		{
			globalScripts[globalName].stop;
			this.changed(\stop, globalName);
		}.defer; // Defer needed for GUI actions from OSC
	}
	*rStopAll { | globalName |
		globalScripts.do { |s| s.stop };
		this.changed(\stopAll);
	}
	*rSet { | globalName, parameter, value |
		globalScripts[globalName].set(parameter, value);
		this.changed(\set, globalName, parameter, value);
	}
	*rAddInput { | globalName, sourceName, input = \in, output = \out |
		globalScripts[globalName].addInput(globalScripts[sourceName], input, output);
		this.changed(\addInput, globalName, sourceName, input, output);
	}
	// not yet implemented!
	*rRemoveInput { | globalName, sourceName, input = \in, output = \out |
		globalScripts[globalName].removeInput(globalScripts[sourceName], input, output);
		this.changed(\removeInput, globalName, globalScripts, input, output);
	}
	add { | scriptName |
		// Access of global scripts by Script[<scriptname>]
		^globalScripts[scriptName];
	}
	*at { | folderName, scriptName |
		// Access of ScriptBrowsers scripts by Script.at(<foldername>, <scriptname>)
		^ScriptBrowser.at(folderName, scriptName);
	}
	*putGlobal { | ... keyValuePairs |
		// "macro" for adding objects to globalParentEnvir
		globalParentEnvir.putPairs(keyValuePairs);
		^keyValuePairs;	// for further use
	}
	*atGlobal { | key |
		// "macro" for getting objects from globalParentEnvir
		^globalParentEnvir[key];
	}
/* { ================= SAVING AND LOADING TO/FROM DISK ================= 
Following methods save / load the script text plus, gui state, parameters (saved values, midiResponders), links to other scripts, audio sample file paths. For plain saving of script text see ScriptBrowser. 
} */
	takeSnapshot {
		// save full current state in snapshots array for saving to disk
		snapshots = snapshots.add(this.makeSnapshotData);
	}
	makeSnapshotData {
		var data;
		^(
			name: name,
			path: path,
			script: script,
			guiBounds: guiBounds,
			id: this.hash, // object id for saving links
			parameterData: parameters collect: _.getSaveData,
			parameterValues: envir[\controlSpecs] collect: envir[_]
			// Following should be encapsulated in parameter data!,
/*			midi: this.getMIDI,
			buffers: this.getBuffers, */
			// But maybe these two will stay:
/*			inputs: this.getInputs,
			outputs: this.getOutputs */
		)
	}
	saveSnapshotDalog {
			CocoaDialog.savePanel({ | argPath |
				this.saveSnapshotOn(argPath);
			}, {
				"save cancelled".postln;
			})
	}
	saveSnapshotOn { | argPath |
		var data;
		Post << "Saving script snapshot:" << argPath << "\n";
		path = argPath;
		this.makeSnapshotData.writeArchive(path);
	}
	loadSnapshotDialog {
		CocoaDialog.getPaths({ | argPaths |
			this.loadSnapshot(argPaths.first);
		}, {
			"load cancelled".postln;
		})
	}
	loadSnapshot { | path |
		var data, pValues;
		data = Object.readArchive(path);
		this.close;
		name = data.name;
		path = data.path;
		script = data.script;
		guiBounds = data.guiBounds;
		id = data.id;
		this.init;
		pValues = data.parameterValues;
		envir[\controlSpecs] do: { |c, i| envir[c[0]] = pValues[i] };
		data.parameterData do: { |p, i| parameters[i].setSaveData(*p) };
	}
	// make yourself playable as event. Requires event types startEvent, stopEvent
	// defined by script Add Script Event Types
	startEvent {
		^(script: this, type: \startScript, delta: 0);
	}
	stopEvent {
		^(script: this, type: \stopScript, delta: 0);
	}
}
