/* (IZ 2005-09-17) {

Class for managing lists of items: Adding, deleting, selecting, performing actions on selected items depending on keyboard-commands, loading and saving. 

This class should be split in two: 
- 1: Class for managing a list
- 2: Class for saving / loading
 
Adapted from Samples class: Load paths from folders into separate events and keep them in a dictionary under the names of the main file name, without the last extension of the name. 
Use: getInfo to get more info about each item and store it in the event, 
	showInfo for displaying the info about each item on an info pane
	testFunc for testing the item
The above 3 variables can be set to customize a SimplePathManager to work for a variety of different objects loaded from file. 
} */

SimplePathManager : Model {
// { 
	// gui:
	var <>name;			// for displaying title on GUI
	var >getInfo;		// function for getting info of an item
	var >showInfo;		// function for displaying info of an item
	var >startTest;	// function for testing an item (play, display ...)
	var >stopTest;		// function for testing an item (play, display ...)
	var <>all;			// the items under their names  
	var <>verbose = false; // if verbose, post each path as it is loaded
	var <window;		// window of GUI for accessing all items
	var <windowRect;	// last position of window for re-making it when closed
	var <selection;	// the currently selected data item
	// UNDER CONSTRUCTION: actions for key strokes on the list view:
	var >spaceFunc, >returnFunc, >enterFunc, >backspaceFunc, >escapeFunc;
// }
	*new { | name = "Items on Disk", getInfo, showInfo, startTest, stopTest |
		^this.newCopyArgs(nil, name, getInfo, showInfo ?? {{ | item, view |
 			view.setString(item.asCompileString, 0, 10000)}},
			startTest, stopTest, ());
	}
	load { | path, loadSubfolders = true, update = true, items |
		/* { Load the full paths of all files matching path. Do not load the 
			contents of the files themselves. Collect the paths and their filename-symbols
			in events. Update or other actions can be done by custom action functions. 
		} */
		var name, newItem;
		items = items ?? { List.new };
		if (verbose) { path.postln; };
		path.pathMatch.do { | p |
			if (loadSubfolders and: { p.last == $/ }) {
				this.load(p ++ "*", loadSubfolders, false, items)
			}{
				newItem = ( path: p, name: name = p.fileSymbol );
				all[name] = newItem;
				items.add(newItem);
			}
		}; // only update at the end, not for each subfolder: 
		if (update) { this.changed(all.keys.asArray.sort, items); };
	}
	makeGui { | rect |
		var pathview, listview, testbutton, load1button, loadFolderButton, infoview, adapter;
		if (window.notNil) { ^window.front };
		window = SCWindow(name,
			rect ? windowRect ?? { Rect(0, 0, 280, 400).fromTop.fromRight });
		window.onClose = {
			windowRect = window.bounds;
			window = nil;
			this.removeDependant(adapter);
		};
		pathview = SCDragSource(window, Rect(3, 3, 274, 17))
			.background_(Color.white(0.5))		// Color(0.93, 0.97, 0.7, 0.3)
			.font_(Font("Helvetica", 9))
			.string_("(path)");
		listview = SCListView(window, Rect(3, 23, 154, 372))
			.items_([])
			.font_(Font("Helvetica", 10));
		loadFolderButton = SCButton(window, Rect(158, 23, 120, 20))
			.states_([["load folder"]])
			.action_ { this.openFolderDialog };
		load1button = SCButton(window, Rect(158, 45, 120, 20))
			.states_([["load file(s)"]])
			.action_ { this.openDialog };
		testbutton = SCButton(window, Rect(158, 67, 120, 20))
			.states_([["test"]])
			.action_ { startTest.(selection) };
		SCButton(window, Rect(158, 89, 120, 20))
			.states_([["save path list..."]])
			.action_ { this.savePathListDialog };
		SCButton(window, Rect(158, 111, 120, 20))
			.states_([["load path list..."]])
			.action_ { this.loadPathListDialog };
		SCButton(window, Rect(158, 133, 120, 20))
			.states_([["save items..."]])
			.action_ { this.saveItemsDialog };
		SCButton(window, Rect(158, 155, 120, 20))
			.states_([["load items..."]])
			.action_ { this.loadItemsDialog };
		infoview = SCTextView(window, Rect(158, 177, 120, 218))
//			.background_(Color(0.5, 0.95, 0.99, 0.3))
//			.stringColor_(Color.yellow)
			.font_(Font("Helvetica", 10));
		listview.action = { | me |
			var name, item, path;
			if (me.value >= me.items.size) { me.value = 0};
			name = me.items[me.value];
			item = all[name.asSymbol];
			selection = nil;
			if (item.notNil) {
				selection = item;
				pathview.object = item.path;
				pathview.string = item.path[item.path.size-60..];
				showInfo.(item, infoview);
			};
			me.focus(true);
		};
		returnFunc = returnFunc ?? {{|argSelection| startTest.(argSelection)}};
		escapeFunc = escapeFunc ?? {{|argSelection| stopTest.(argSelection)}};
		enterFunc = enterFunc ?? {
			{|argSelection| if (selection.notNil) {selection.path.postln}}
		};
		backspaceFunc = backspaceFunc ?? {{ this.deleteItem }};
		listview.keyDownAction = { | v, char, mod, unic, key |
			// customizeable actions for space, return, enter, backspace, escape keys
			switch (char,
				// space
				$ , { spaceFunc.(selection) },
				// return: test selection
				13.asAscii, { returnFunc.(selection) },
				// escape: stop any playing tests
				27.asAscii, { escapeFunc.(selection) },
				// enter: post path
				3.asAscii, { enterFunc.(selection) },
				// backspace: delete item
				127.asAscii,{ backspaceFunc.(selection) },
				{ v.defaultKeyDownAction(char, mod, unic, key) }
			);
		};
		adapter = { | who, how, items |
			switch (how,
				\items, {
					listview.items = items ? [];
				}
			);
			listview.doAction;
		};
		this.addDependant(adapter);
		window.front;
		this.changed; // update views with existing contents
	}
	openFolderDialog {
// { load all files in a folder and its subfolders. 
//	Here using savePanel dialog as a work-around for selecting a folder "f" even if that folder only contains subfolders. In that case, it would not be possible to select folder "f" with CocoaDialog.getPaths!
//	} 
		CocoaDialog.savePanel({ | path |
			this.load(path.dirname ++ "/*");
			}, {
					"load cancelled".postln;
			});
	}
	openDialog {
		CocoaDialog.getPaths({ | paths |
			paths.do { |p| this.load(p) };
		}, {
			"load cancelled".postln;
		})
	}
	deleteItem {
		var key;
		if (selection.isNil) { ^this };
		key = all.findKeyForValue(selection);
		Post << "Deleting: " << key << "\n";
		all.removeAt(key);
		selection = nil; // if no gui, selection becomes nil
		this.changed(\items, all.keys.asArray.sort); // otherwise new selection from listview
	}
	clear {
		all = ();
		selection = nil;
		Post << name << " - deleted all items\n";
		this.changed(\items, []);
	}

	// saving and loading current lists of paths
	savePathListDialog {
		CocoaDialog.savePanel({ | path |
			this.savePathList(path);
		}, {
			"load cancelled".postln;
		})
	}
	savePathList { | path |
		// saving all paths. One path per line, for easy editing
		var file;
		file = File(path, "w");
		(all.values.asArray collect: _.path).sort do: { |p|
			file.write(p).write("\n");
		};
		file.close;
	}
	loadPathListDialog {
		CocoaDialog.getPaths({ | paths |
			this.loadPathList(paths.first);
		}, {
			"load cancelled".postln;
		})
	}
	loadPathList { | path |
		var paths;
		paths = File(path, "r").readAllString.split($\n);
		paths.postln;
	}
	// saving and loading current item data
	saveItemsDialog {
		CocoaDialog.savePanel({ | path |
			this.saveItems(path);
		}, {
			"load cancelled".postln;
		})
	}
	saveItems { | path |
		var file;
		all.values.asArray.writeArchive(path);
	}
	loadItemsDialog {
		CocoaDialog.getPaths({ | paths |
			this.loadItems(paths.first);
		}, {
			"load cancelled".postln;
		})
	}
	loadItems { | path |
		var items;
		items = Object.readArchive(path);
		items.do { | i | all[i.name] = i };
		this.changed(all.keys.asArray.sort);
	}
	at { | ... indexArray |
		if (indexArray.size == 1) {
			^all[indexArray.first]
		}{
			^indexArray collect: all[_]
		}
	}
	put { | argName, argItem |
		all[argName] = argItem;
		this.changed(\items, all.keys.asArray.sort);
	}
	getInfo { | item |
		^getInfo.(item);
	}
}
